﻿Imports System.Collections.ObjectModel
Imports System.Globalization
Imports System.Linq
Imports System.ServiceModel
Imports Microsoft.Xrm.Sdk

''' <summary>
''' Base class for all Plugins.
''' </summary>    
Public Class Plugin
    Implements IPlugin
    Protected Class LocalPluginContext
        Friend Property ServiceProvider() As IServiceProvider
            Get
                Return m_ServiceProvider
            End Get

            Private Set(value As IServiceProvider)
                m_ServiceProvider = Value
            End Set
        End Property
        Private m_ServiceProvider As IServiceProvider

        Friend Property OrganizationService() As IOrganizationService
            Get
                Return m_OrganizationService
            End Get

            Private Set(value As IOrganizationService)
                m_OrganizationService = Value
            End Set
        End Property
        Private m_OrganizationService As IOrganizationService

        Friend Property PluginExecutionContext() As IPluginExecutionContext
            Get
                Return m_PluginExecutionContext
            End Get

            Private Set(value As IPluginExecutionContext)
                m_PluginExecutionContext = Value
            End Set
        End Property
        Private m_PluginExecutionContext As IPluginExecutionContext

        Friend Property TracingService() As ITracingService
            Get
                Return m_TracingService
            End Get

            Private Set(value As ITracingService)
                m_TracingService = Value
            End Set
        End Property
        Private m_TracingService As ITracingService

        Private Sub New()
        End Sub

        Friend Sub New(serviceProvider As IServiceProvider)
            If serviceProvider Is Nothing Then
                Throw New ArgumentNullException("serviceProvider")
            End If

            ' Obtain the execution context service from the service provider.
            Me.PluginExecutionContext = DirectCast(serviceProvider.GetService(GetType(IPluginExecutionContext)), IPluginExecutionContext)

            ' Obtain the tracing service from the service provider.
            Me.TracingService = DirectCast(serviceProvider.GetService(GetType(ITracingService)), ITracingService)

            ' Obtain the Organization Service factory service from the service provider
            Dim factory As IOrganizationServiceFactory = DirectCast(serviceProvider.GetService(GetType(IOrganizationServiceFactory)), IOrganizationServiceFactory)

            ' Use the factory to generate the Organization Service.
            Me.OrganizationService = factory.CreateOrganizationService(Me.PluginExecutionContext.UserId)
        End Sub

        Friend Sub Trace(message As String)
            If String.IsNullOrWhiteSpace(message) OrElse Me.TracingService Is Nothing Then
                Return
            End If

            If Me.PluginExecutionContext Is Nothing Then
                Me.TracingService.Trace(message)
            Else
                Me.TracingService.Trace("{0}, Correlation Id: {1}, Initiating User: {2}", message, Me.PluginExecutionContext.CorrelationId, Me.PluginExecutionContext.InitiatingUserId)
            End If
        End Sub
    End Class

    Private m_registeredEvents As Collection(Of Tuple(Of Integer, String, String, Action(Of LocalPluginContext)))

    ''' <summary>
    ''' Gets the List of events that the plug-in should fire for. Each List
    ''' Item is a <see cref="System.Tuple"/> containing the Pipeline Stage, Message and (optionally) the Primary Entity. 
    ''' In addition, the fourth parameter provide the delegate to invoke on a matching registration.
    ''' </summary>
    Protected ReadOnly Property RegisteredEvents() As Collection(Of Tuple(Of Integer, String, String, Action(Of LocalPluginContext)))
        Get
            If Me.m_registeredEvents Is Nothing Then
                Me.m_registeredEvents = New Collection(Of Tuple(Of Integer, String, String, Action(Of LocalPluginContext)))()
            End If

            Return Me.m_registeredEvents
        End Get
    End Property

    ''' <summary>
    ''' Gets or sets the name of the child class.
    ''' </summary>
    ''' <value>The name of the child class.</value>
    Protected Property ChildClassName() As String
        Get
            Return m_ChildClassName
        End Get

        Private Set(value As String)
            m_ChildClassName = Value
        End Set
    End Property
    Private m_ChildClassName As String

    ''' <summary>
    ''' Initializes a new instance of the <see cref="Plugin"/> class.
    ''' </summary>
    ''' <param name="childClassName">The <see cref=" cred="Type"/> of the derived class.</param>
    Friend Sub New(childClassName As Type)
        Me.ChildClassName = childClassName.ToString()
    End Sub

    ''' <summary>
    ''' Executes the plug-in.
    ''' </summary>
    ''' <param name="serviceProvider">The service provider.</param>
    ''' <remarks>
    ''' For improved performance, Microsoft Dynamics CRM caches plug-in instances. 
    ''' The plug-in's Execute method should be written to be stateless as the constructor 
    ''' is not called for every invocation of the plug-in. Also, multiple system threads 
    ''' could execute the plug-in at the same time. All per invocation state information 
    ''' is stored in the context. This means that you should not use global variables in plug-ins.
    ''' </remarks>
    Public Sub Execute(serviceProvider As IServiceProvider) Implements IPlugin.Execute
        If serviceProvider Is Nothing Then
            Throw New ArgumentNullException("serviceProvider")
        End If

        ' Construct the Local plug-in context.
        Dim localcontext As New LocalPluginContext(serviceProvider)

        localcontext.Trace(String.Format(CultureInfo.InvariantCulture, "Entered {0}.Execute()", Me.ChildClassName))

        Try
            ' Iterate over all of the expected registered events to ensure that the plugin
            ' has been invoked by an expected event
            ' For any given plug-in event at an instance in time, we would expect at most 1 result to match.
            Dim entityAction As Action(Of LocalPluginContext) = (From a In Me.RegisteredEvents Where (a.Item1 = localcontext.PluginExecutionContext.Stage AndAlso a.Item2 = localcontext.PluginExecutionContext.MessageName AndAlso (IIf(String.IsNullOrWhiteSpace(a.Item3), True, a.Item3 = localcontext.PluginExecutionContext.PrimaryEntityName))) Select a.Item4).FirstOrDefault()

            If entityAction IsNot Nothing Then
                localcontext.Trace(String.Format(CultureInfo.InvariantCulture, "{0} is firing for Entity: {1}, Message: {2}", Me.ChildClassName, localcontext.PluginExecutionContext.PrimaryEntityName, localcontext.PluginExecutionContext.MessageName))

                entityAction.Invoke(localcontext)

                ' now exit - if the derived plug-in has incorrectly registered overlapping event registrations,
                ' guard against multiple executions.
                Return
            End If
        Catch e As FaultException
            localcontext.Trace(String.Format(CultureInfo.InvariantCulture, "Exception: {0}", e.ToString()))

            ' Handle the exception.
            Throw
        Finally
            localcontext.Trace(String.Format(CultureInfo.InvariantCulture, "Exiting {0}.Execute()", Me.ChildClassName))
        End Try
    End Sub
End Class
